@@ -1,11 +1,22 @@
 use std::convert::TryFrom;
 
+use snafu::prelude::*;
+
 pub mod fs;
 pub mod http;
 pub mod memory;
 pub mod object_store;
 
-use crate::error::ColumnQError;
+#[derive(Debug, Snafu)]
+pub enum Error {
+    #[snafu(display("Generic {} error: {}", backend, source))]
+    Generic {
+        backend: &'static str,
+        source: Box<dyn std::error::Error + Send + Sync + 'static>,
+    },
+    #[snafu(display("Unsupported URI scheme: {scheme}"))]
+    InvalidUriScheme { scheme: String },
+}
 
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub enum BlobStoreType {
@@ -18,7 +29,7 @@ pub enum BlobStoreType {
 }
 
 impl TryFrom<Option<&uriparse::Scheme<'_>>> for BlobStoreType {
-    type Error = ColumnQError;
+    type Error = Error;
 
     fn try_from(scheme: Option<&uriparse::Scheme<'_>>) -> Result<Self, Self::Error> {
         match scheme {
@@ -28,16 +39,15 @@ impl TryFrom<Option<&uriparse::Scheme<'_>>> for BlobStoreType {
             }
             Some(uriparse::Scheme::HTTP) | Some(uriparse::Scheme::HTTPS) => Ok(BlobStoreType::Http),
             Some(uriparse::Scheme::Unregistered(s)) => BlobStoreType::try_from(s.as_str()),
-            _ => Err(ColumnQError::InvalidUri(format!(
-                "Unsupported scheme: {:?}",
-                scheme
-            ))),
+            _ => Err(Error::InvalidUriScheme {
+                scheme: format!("{:?}", scheme),
+            }),
         }
     }
 }
 
 impl TryFrom<&str> for BlobStoreType {
-    type Error = ColumnQError;
+    type Error = Error;
 
     fn try_from(scheme: &str) -> Result<Self, Self::Error> {
         match scheme {
@@ -45,10 +55,9 @@ impl TryFrom<&str> for BlobStoreType {
             "gs" => Ok(BlobStoreType::GCS),
             "az" | "adl" | "adfs" | "adfss" | "azure" => Ok(BlobStoreType::Azure),
             "memory" => Ok(BlobStoreType::Memory),
-            _ => Err(ColumnQError::InvalidUri(format!(
-                "Unsupported scheme: {:?}",
-                scheme
-            ))),
+            _ => Err(Error::InvalidUriScheme {
+                scheme: scheme.to_string(),
+            }),
         }
     }
 }
