@@ -2,38 +2,61 @@ use std::io::{BufReader, Read};
 use std::sync::Arc;
 
 use datafusion::arrow::datatypes::{Schema, SchemaRef};
-use datafusion::arrow::json::reader::{infer_json_schema, Decoder, DecoderOptions, ValueIter};
+#[allow(deprecated)]
+use datafusion::arrow::json::reader::{infer_json_schema, ReaderBuilder};
 use datafusion::arrow::record_batch::RecordBatch;
+use snafu::prelude::*;
 
-use crate::error::ColumnQError;
-use crate::table::TableSource;
+use crate::table::{self, TableSource};
 
-fn json_schema_from_reader<R: Read>(r: R) -> Result<Schema, ColumnQError> {
+#[derive(Debug, Snafu)]
+pub enum Error {
+    #[snafu(display("Failed to infer NDJSON schema: {source}"))]
+    InferSchema {
+        source: datafusion::arrow::error::ArrowError,
+    },
+    #[snafu(display("Found empty NDJSON schema"))]
+    EmptySchema {},
+    #[snafu(display("Failed to build arrow reader: {source}"))]
+    BuildReader {
+        source: datafusion::arrow::error::ArrowError,
+    },
+    #[snafu(display("Failed to collect arrow record batches: {source}"))]
+    CollectBatches {
+        source: datafusion::arrow::error::ArrowError,
+    },
+}
+
+fn json_schema_from_reader<R: Read>(r: R) -> Result<Schema, table::Error> {
     let mut reader = BufReader::new(r);
-    Ok(infer_json_schema(&mut reader, None)?)
+    infer_json_schema(&mut reader, None)
+        .context(InferSchemaSnafu)
+        .context(table::LoadNdJsonSnafu)
 }
 
 fn decode_json_from_reader<R: Read>(
     r: R,
     schema_ref: SchemaRef,
     batch_size: usize,
-) -> Result<Vec<RecordBatch>, ColumnQError> {
-    let decoder = Decoder::new(
-        schema_ref,
-        DecoderOptions::new().with_batch_size(batch_size),
-    );
-    let mut reader = BufReader::new(r);
-    let mut value_reader = ValueIter::new(&mut reader, None);
-    let mut batches = vec![];
-    while let Some(batch) = decoder.next_batch(&mut value_reader)? {
-        batches.push(batch);
-    }
+) -> Result<Vec<RecordBatch>, table::Error> {
+    let batch_reader = ReaderBuilder::new(schema_ref)
+        .with_batch_size(batch_size)
+        .build(BufReader::new(r))
+        .context(BuildReaderSnafu)
+        .context(table::LoadNdJsonSnafu)?;
+
+    let batches = batch_reader
+        .collect::<Result<Vec<RecordBatch>, _>>()
+        .context(CollectBatchesSnafu)
+        .context(table::LoadNdJsonSnafu)?;
+
     Ok(batches)
 }
 
 pub async fn to_mem_table(
     t: &TableSource,
-) -> Result<datafusion::datasource::MemTable, ColumnQError> {
+    dfctx: &datafusion::execution::context::SessionContext,
+) -> Result<datafusion::datasource::MemTable, table::Error> {
     let batch_size = t.batch_size;
 
     let schema_ref: SchemaRef = match &t.schema {
@@ -41,39 +64,47 @@ pub async fn to_mem_table(
         // infer schema from data if not provided by user
         None => {
             let inferred_schema: Vec<Schema> =
-                partitions_from_table_source!(t, json_schema_from_reader)?;
+                partitions_from_table_source!(t, json_schema_from_reader, dfctx)
+                    .context(table::IoSnafu)?;
             if inferred_schema.is_empty() {
-                return Err(ColumnQError::LoadJson("failed to load schema".to_string()));
+                return Err(Error::EmptySchema {}).context(table::LoadNdJsonSnafu);
             }
-            Arc::new(Schema::try_merge(inferred_schema)?)
+            Arc::new(
+                Schema::try_merge(inferred_schema)
+                    .context(InferSchemaSnafu)
+                    .context(table::LoadNdJsonSnafu)?,
+            )
         }
     };
 
-    let partitions: Vec<Vec<RecordBatch>> =
-        partitions_from_table_source!(t, |reader| decode_json_from_reader(
-            reader,
-            schema_ref.clone(),
-            batch_size
-        ))?;
+    let partitions: Vec<Vec<RecordBatch>> = partitions_from_table_source!(
+        t,
+        |reader| decode_json_from_reader(reader, schema_ref.clone(), batch_size),
+        dfctx
+    )
+    .context(table::IoSnafu)?;
 
-    Ok(datafusion::datasource::MemTable::try_new(
-        schema_ref, partitions,
-    )?)
+    datafusion::datasource::MemTable::try_new(schema_ref, partitions)
+        .context(table::CreateMemTableSnafu)
 }
 
 #[cfg(test)]
 mod tests {
-    use datafusion::datasource::TableProvider;
+    use datafusion::{datasource::TableProvider, prelude::SessionContext};
 
     use super::*;
     use crate::{table::TableSource, test_util::test_data_path};
 
     #[tokio::test]
     async fn load_simple_ndjson_file() {
-        let t = to_mem_table(&TableSource::new(
-            "spacex_launches".to_string(),
-            test_data_path("spacex_launches.ndjson"),
-        ))
+        let ctx = SessionContext::new();
+        let t = to_mem_table(
+            &TableSource::new(
+                "spacex_launches".to_string(),
+                test_data_path("spacex_launches.ndjson"),
+            ),
+            &ctx,
+        )
         .await
         .unwrap();
 
@@ -118,10 +149,14 @@ mod tests {
 
     #[tokio::test]
     async fn load_simple_jsonl_file() {
-        let t = to_mem_table(&TableSource::new(
-            "spacex_launches".to_string(),
-            test_data_path("spacex_launches.jsonl"),
-        ))
+        let ctx = SessionContext::new();
+        let t = to_mem_table(
+            &TableSource::new(
+                "spacex_launches".to_string(),
+                test_data_path("spacex_launches.jsonl"),
+            ),
+            &ctx,
+        )
         .await
         .unwrap();
 
