@@ -13,6 +13,7 @@ use convergence::protocol::{ErrorResponse, FieldDescription, SqlState};
 use convergence::protocol_ext::DataRowBatch;
 use convergence_arrow::table::{record_batch_to_rows, schema_to_field_desc};
 use log::info;
+use snafu::{whatever, Whatever};
 use tokio::net::TcpListener;
 
 use crate::config::Config;
@@ -20,18 +21,19 @@ use crate::context::RoapiContext;
 use crate::server::RunnableServer;
 
 fn df_err_to_sql(err: DataFusionError) -> ErrorResponse {
-    ErrorResponse::error(SqlState::DATA_EXCEPTION, err.to_string())
+    ErrorResponse::error(SqlState::DataException, err.to_string())
 }
 
 /// A portal built using a logical DataFusion query plan.
 pub struct DataFusionPortal {
-    df: Arc<DataFrame>,
+    df: DataFrame,
 }
 
 #[async_trait]
 impl Portal for DataFusionPortal {
     async fn fetch(&mut self, batch: &mut DataRowBatch) -> Result<(), ErrorResponse> {
-        for arrow_batch in self.df.collect().await.map_err(df_err_to_sql)? {
+        let arrow_batches = self.df.clone().collect().await.map_err(df_err_to_sql)?;
+        for arrow_batch in arrow_batches {
             record_batch_to_rows(&arrow_batch, batch)?;
         }
         Ok(())
@@ -42,6 +44,26 @@ pub struct RoapiContextEngine<H: RoapiContext> {
     pub ctx: Arc<H>,
 }
 
+impl<H: RoapiContext> RoapiContextEngine<H> {
+    fn ignored_statement(statement: &Statement) -> bool {
+        !matches!(
+            statement,
+            Statement::Query { .. }
+                | Statement::Analyze { .. }
+                | Statement::Fetch { .. }
+                | Statement::ShowFunctions { .. }
+                | Statement::ShowVariable { .. }
+                | Statement::ShowVariables { .. }
+                | Statement::ShowCollation { .. }
+                | Statement::Assert { .. }
+                | Statement::ExplainTable { .. }
+                | Statement::Explain { .. }
+                | Statement::ShowColumns { .. }
+                | Statement::ShowTables { .. }
+        )
+    }
+}
+
 #[async_trait]
 impl<H: RoapiContext> Engine for RoapiContextEngine<H> {
     type PortalType = DataFusionPortal;
@@ -50,6 +72,9 @@ impl<H: RoapiContext> Engine for RoapiContextEngine<H> {
         &mut self,
         statement: &Statement,
     ) -> Result<Vec<FieldDescription>, ErrorResponse> {
+        if RoapiContextEngine::<H>::ignored_statement(statement) {
+            return Ok(vec![]);
+        }
         let query = statement.to_string();
         info!("preparing query: {}", &query);
         let df = self.ctx.sql_to_df(&query).await.map_err(df_err_to_sql)?;
@@ -60,9 +85,19 @@ impl<H: RoapiContext> Engine for RoapiContextEngine<H> {
         &mut self,
         statement: &Statement,
     ) -> Result<Self::PortalType, ErrorResponse> {
-        let query = statement.to_string();
-        let df = self.ctx.sql_to_df(&query).await.map_err(df_err_to_sql)?;
-        Ok(DataFusionPortal { df })
+        if RoapiContextEngine::<H>::ignored_statement(statement) {
+            Ok(DataFusionPortal {
+                df: self
+                    .ctx
+                    .sql_to_df("SELECT 1 WHERE 1 = 2")
+                    .await
+                    .map_err(df_err_to_sql)?,
+            })
+        } else {
+            let query = statement.to_string();
+            let df = self.ctx.sql_to_df(&query).await.map_err(df_err_to_sql)?;
+            Ok(DataFusionPortal { df })
+        }
     }
 }
 
@@ -74,7 +109,7 @@ pub struct PostgresServer<H: RoapiContext> {
 
 impl<H: RoapiContext> PostgresServer<H> {
     pub async fn new(ctx: Arc<H>, config: &Config, default_host: String) -> Self {
-        let default_addr = format!("{}:5432", default_host);
+        let default_addr = format!("{default_host}:5432");
 
         let addr = config
             .addr
@@ -101,11 +136,14 @@ impl<H: RoapiContext> RunnableServer for PostgresServer<H> {
         self.addr
     }
 
-    async fn run(&self) -> anyhow::Result<()> {
+    async fn run(&self) -> Result<(), Whatever> {
         use convergence::connection::Connection;
 
         loop {
-            let (stream, _) = self.listener.accept().await?;
+            let (stream, _) = whatever!(
+                self.listener.accept().await,
+                "Failed to create postgres TCP listener"
+            );
             let engine = RoapiContextEngine {
                 ctx: self.ctx.clone(),
             };
